// db.h: interface for the db class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_DB_H__4360E5C1_FA02_11D5_AB34_8D852566F660__INCLUDED_)
#define AFX_DB_H__4360E5C1_FA02_11D5_AB34_8D852566F660__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#pragma warning(disable : 4786)

#include "assert.h"
#include "comdef.h"
#include <map>
#include <vector>
#include <string>

namespace svp
{

///////////////////////////////////////////////////////////////////////////////////
// Definition

#ifdef _UNICODE
typedef std::wstring cdb_tstring;
#else
typedef std::string cdb_tstring;
#endif

// please use this struct as base for you db-struct
struct cdb_base_struct
{
	virtual cdb_tstring GetKey(long lKeyNumber) const = 0;
};

// "template" class for key generation
class cdb_keygen
{
public:

	static cdb_tstring GetKey(long l);
	static cdb_tstring GetKey(std::string s, long lMaxKeyLen);
	static cdb_tstring GetKey(std::wstring s, long lMaxKeyLen);
	static cdb_tstring GetKey(LPCTSTR pString, long lMaxKeyLen);
};

// database template class
template<class T>
class cdb
{

protected:

	bool m_bReBuildKeys;

	typedef std::map<cdb_tstring, long> key_type;

	typedef std::map<cdb_tstring, long> synonyms_count_type;

	typedef std::vector<key_type> key_map_vector_type;

	key_map_vector_type m_Key;

	typedef std::vector<T> data_type;

	data_type m_Data;

	synonyms_count_type m_SynonymsCount;

	long m_lKeys;

	void put_to_map(const T& x, long lPos);

	void clear_keys();

	void rebuild_keys();

public:

	enum { ITEM_NOT_FOUND = -1 };

// all

	/** Create, lkeys is number of keys. */
	cdb(long lKeys = 0);

	/** Push item to front. */
	void push_front(const T& x);

	/** Push item to back. */
	void push_back(const T& x);

	/** Insert item at position. */
	void insert_at(long lPosition, const T& x);

	/** Erase item at position. */
	void erase_at(long lPosition);

// access as vector

	/** Function return numver of items. */
	long size();

	/** Access to item on defined position. */
	T& operator[](long l);

// access by key

	/** Find item.*/
	bool find(long lKeyNumber, const T& x, T& res, long lSyn = 0);

	/** Find item, return item position. */
	long get_position(long lKeyNumber, const T& x, long lSyn = 0);

	/** Erase specified item. */
	bool erase(long lKeyNumber, const T& x, long lSyn = 0);

	/** Get count of synonyms for specified item. */
	long synonyms_count(long lKeyNumber, const T& x);
};

///////////////////////////////////////////////////////////////////////////////////
// Template implementation

template<class T>
long cdb<T>::size()
{
	return m_Data.size();
}

template<class T>
T& cdb<T>::operator[](long l)
{
	assert(l >= 0 && l < m_Data.size());

	return m_Data[l];
}

template<class T>
long cdb<T>::synonyms_count(long lKeyNumber, const T& x)
{
	assert(lKeyNumber >= 0 && lKeyNumber < m_lKeys);

	return m_SynonymsCount[x.GetKey(lKeyNumber)];
}

template<class T>
void cdb<T>::rebuild_keys()
{
	if (m_bReBuildKeys)
	{
		clear_keys();

		for (long k = 0; k < m_Data.size(); k++)
		{
			put_to_map(m_Data[k], k);
		}

		m_bReBuildKeys = false;
	}
}

template<class T>
long cdb<T>::get_position(long lKeyNumber, const T& x, long lSyn /* = 0 */)
{
	assert(lKeyNumber >= 0 && lKeyNumber < m_lKeys);

	rebuild_keys();

	if (m_Key[lKeyNumber].find(x.GetKey(lKeyNumber) + cdb_keygen::GetKey(lSyn)) != m_Key[lKeyNumber].end())
	{
		return m_Key[lKeyNumber][x.GetKey(lKeyNumber) + cdb_keygen::GetKey(lSyn)]; 
	}
	else
	{
		return ITEM_NOT_FOUND;
	}
}

template<class T>
bool cdb<T>::find(long lKeyNumber, const T& x, T& res, long lSyn /* = 0 */)
{
	assert(lKeyNumber >= 0 && lKeyNumber < m_lKeys);

	rebuild_keys();

	if (m_Key[lKeyNumber].find(x.GetKey(lKeyNumber) + cdb_keygen::GetKey(lSyn)) != m_Key[lKeyNumber].end())
	{
		res = m_Data[m_Key[lKeyNumber][x.GetKey(lKeyNumber) + cdb_keygen::GetKey(lSyn)]];
		return true; 
	}
	else
	{
		return false;
	}
}

template<class T>
void cdb<T>::put_to_map(const T& x, long lPos)
{
	for (long k = 0; k < m_Key.size(); k++)
	{
		long dbl;
		if (m_SynonymsCount.find(x.GetKey(k)) == m_SynonymsCount.end())
		{
			dbl = 0;
			m_SynonymsCount[x.GetKey(k)] = 1;
		}
		else
		{
			dbl = m_SynonymsCount[x.GetKey(k)]++;
		}

		m_Key[k][x.GetKey(k) + cdb_keygen::GetKey(dbl)] = lPos;
	}
}

template<class T>
void cdb<T>::push_back(const T& x)
{
	if (!m_bReBuildKeys)
	{
		put_to_map(x, m_Data.size());
	}

	m_Data.push_back(x);
}

template<class T>
void cdb<T>::push_front(const T& x)
{
	insert_at(0, x);
}

template<class T>
void cdb<T>::clear_keys()
{
	for (long k = 0; k < m_lKeys; k++)
	{
		m_Key[k].clear();
	}

	m_SynonymsCount.clear();
}

template<class T>
void cdb<T>::insert_at(long lPosition, const T& x)
{
	assert(lPosition >= 0 && lPosition < m_Data.size());

	m_Data.insert(m_Data.begin() + lPosition, x);

	m_bReBuildKeys = true;
}

template<class T>
bool cdb<T>::erase(long lKeyNumber, const T& x, long lSyn /* = 0 */)
{
	// key number i sinvalid
	assert(lKeyNumber >= 0 && lKeyNumber < m_lKeys);

	long lPos = get_position(lKeyNumber, x, lSyn);

	if (lPos == ITEM_NOT_FOUND)
	{
		return false;
	}
	else
	{
		erase_at(lPos);
		return true;
	}
}

template<class T>
void cdb<T>::erase_at(long lPosition)
{
	// bad delete position
	assert(lPosition >= 0 && lPosition < m_Data.size());

	m_Data.erase(m_Data.begin() + lPosition);

	m_bReBuildKeys = true;
}

template<class T>
cdb<T>::cdb(long lKeys /* = 0 */)
{
	m_lKeys = lKeys;

	for (long k = 0; k < m_lKeys; k++)
	{
		m_Key.push_back(key_type());
	}

	m_bReBuildKeys = false;
}

}
#endif // !defined(AFX_DB_H__4360E5C1_FA02_11D5_AB34_8D852566F660__INCLUDED_)
